/* 
 * Copyright (C) 2005  Christian Cier-Zniewski
 * Author: Christian Cier-Zniewski <c.cier@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <pthread.h>
#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <memory.h>
#include <netdb.h>
#include <stdlib.h>

#include "global.h"

/*
 * SetNonBlocking sets a socket into non-blocking mode.
 */


int
set_non_blocking(int sock)
{
   if (fcntl(sock, F_SETFL, O_NONBLOCK) < 0)
   {
      fprintf(stderr, "Setting socket to non-blocking failed.\n");
      return -1;
   }
   return 1;
}

// Establish tcp connection to server
// RETURN: socket handle
int 
tcp_connect_to_server (char *host, int port)
{
   struct hostent *he=NULL;
   int one=1;
   struct sockaddr_in s;
   int sock;

   if ( (sock = socket (AF_INET, SOCK_STREAM, 0)) <0) 
   {
      fprintf(stderr, "Error creating communication socket(Control): %d\n", errno);
      exit(2); 
   }

  
   /* if the server wasnt specified as an ip address, look it up */
   if (!inet_aton(host, &s.sin_addr)) 
   {
      if ( (he = gethostbyname(host)) )
	 memcpy (&s.sin_addr.s_addr, he->h_addr, he->h_length);
      else
      {
	 fprintf(stderr, "Couldnt resolve host!\n");
         close(sock);
	 return (0);
      }
   }
   
   s.sin_port = htons(port);
   s.sin_family = AF_INET;

   if (connect(sock,(struct sockaddr*) &s, sizeof(s)) < 0)
   {
      fprintf(stderr, "Connect error\n");
      close(sock);
      return (0);
   }
   if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *)&one, sizeof(one)) < 0)
   {
      fprintf(stderr, "Error setting socket options\n");
      close(sock);
      return (0);
   }

   if (!set_non_blocking(sock)) return -1;
  
   return (sock);
}

int 
tcp_putdata(int sock, unsigned char *buf, unsigned int len)
{	
   int numbytes;
   
   if ((numbytes=send(sock, buf, len, 0)) <0)
   {
      perror("recv:");
      return(-1);
   }
   return numbytes;
}

int
tcp_getdata(int sock, unsigned char *buf, unsigned int len)
{
   int numbytes;
   fd_set read_fds;   // master file descriptor list
   struct timeval timeout;  /* Timeout for select */

   FD_ZERO(&read_fds);
   FD_SET(sock, &read_fds);
   timeout.tv_sec = 0;
   timeout.tv_usec = 0;

   if (select(sock+1, &read_fds, NULL, NULL, &timeout) <0) 
   {
      perror("select");
      exit(1);
   }
   
   if (FD_ISSET(sock, &read_fds)) 
   {
      if ((numbytes=recv(sock, buf, len, 0)) <0)
      {
        perror("recv:");
      	return(-1);
      }

#if 0   
      int i;

      fprintf(stderr,"TCP: %d bytes received.\n Data: ",numbytes);
      for (i=0;i<numbytes;i++)
      {
         fprintf(stderr,"%02X ", *(buf+i));
      }
      fprintf(stderr,"\n");
#endif      
      
      return(numbytes);	
   } 
   else return(0);
   
}

int
udp_bind(int port)
{
   struct sockaddr_in s;
   int sock;

   if ( (sock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) <0) 
   {
      fprintf(stderr, "Error creating communication socket(TSreceiver): %d\n", errno);
      exit(2); 
   }
   
   s.sin_family=AF_INET;
   s.sin_addr.s_addr=htonl(INADDR_ANY);
   s.sin_port=htons(port);
   
   if( (bind(sock,(struct sockaddr*)&s,sizeof(s))) <0)
   {
   	fprintf(stderr, "Error binding socket\n");
   	close(sock);
   	exit(2);
   }
   
   if (!set_non_blocking(sock)) return -1;

   return(sock);
}

int
udp_getdata(int sock, unsigned char *buf, unsigned int len)
{
   struct sockaddr_in their_addr;
   int addr_len, numbytes;
   fd_set read_fds;   // master file descriptor list
   struct timeval timeout;  /* Timeout for select */

   FD_ZERO(&read_fds);
   FD_SET(sock, &read_fds);
   timeout.tv_sec = 0;
   timeout.tv_usec = 0;

   if (select(sock+1, &read_fds, NULL, NULL, &timeout) <0) 
   {
      perror("select");
      exit(1);
   }
   
   if (FD_ISSET(sock, &read_fds)) 
   {
      addr_len = sizeof(struct sockaddr);
      if ((numbytes=recvfrom(sock, buf, len, 0, (struct sockaddr *)&their_addr, &addr_len)) <0)
      {
        perror("recvfrom:");
      	return(-1);
      }
   
      fprintf(stderr,"got packet from %s\n",inet_ntoa(their_addr.sin_addr));
      int i;
      fprintf(stderr,"%d bytes received.\n Data: ",numbytes);
      for (i=0;i<numbytes;i++)
      {
         fprintf(stderr,"%02X ", *(buf+i));
      }
      fprintf(stderr,"\n");
      return(numbytes);	
   } 
   else return(0);
}

