/* 
 * Copyright (C) 2005  Christian Cier-Zniewski
 * Author: Christian Cier-Zniewski <c.cier@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <pthread.h>
#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <memory.h>
#include <netdb.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <linux/fb.h>


#include "global.h"
#include "rfbproto.h"

#define MAXRECVSIZE 7*188
#define MAXCOLORS 256

static volatile int OSDreceiver_active;
pthread_t OSDreceiver_thread;
unsigned char OSDbuffer[MAXRECVSIZE];

struct fb_var_screeninfo screeninfo, oldscreen;
int fb;
unsigned char *lfb;
int rcfd;

__u16 red[MAXCOLORS], green[MAXCOLORS], blue[MAXCOLORS], alpha[MAXCOLORS];
unsigned char ramp[MAXCOLORS], rampalpha[MAXCOLORS];


int 
setmode(unsigned int nxRes, unsigned int nyRes, unsigned int nbpp)
{
        screeninfo.xres_virtual=screeninfo.xres=nxRes;
        screeninfo.yres_virtual=screeninfo.yres=nyRes;
        screeninfo.height=0;
        screeninfo.width=0;
        screeninfo.xoffset=screeninfo.yoffset=0;
        screeninfo.bits_per_pixel=nbpp;
        if (ioctl(fb, FBIOPUT_VSCREENINFO, &screeninfo)<0)
        {
                perror("FBIOPUT_VSCREENINFO");
                printf("fb failed\n");
                return -1;
        }
        if ((screeninfo.xres!=nxRes) && (screeninfo.yres!=nyRes) && (screeninfo.bits_per_pixel!=nbpp))
        {
                printf("SetMode failed: wanted: %dx%dx%d, got %dx%dx%d\r\n",
                        nxRes, nyRes, nbpp, screeninfo.xres, screeninfo.yres, screeninfo.bits_per_pixel);
        }
        struct fb_fix_screeninfo fix;
        if (ioctl(fb, FBIOGET_FSCREENINFO, &fix)<0)
        {
                perror("FBIOGET_FSCREENINFO");
                printf("fb failed\n");
        }
        fprintf(stderr,"stride:%d\n",fix.line_length);
        memset(lfb, 0, fix.line_length*screeninfo.yres);
        return 1;
}

int 
fb_init(void)
{
    unsigned int available=0;
    int i,d;
    int gamma=128;
    int brightness=128;
    int alpha=255;
        
        fb=open("/dev/fb/0", O_RDWR);
        if (fb<0)
        {
                perror("/dev/fb/0");
                goto nolfb;
        }
        if (ioctl(fb, FBIOGET_VSCREENINFO, &screeninfo)<0)
        {
                perror("FBIOGET_VSCREENINFO");
                goto nolfb;
        }
        memcpy(&oldscreen, &screeninfo, sizeof(screeninfo));

        struct fb_fix_screeninfo fix;

        if (ioctl(fb, FBIOGET_FSCREENINFO, &fix)<0)
        {
                perror("FBIOGET_FSCREENINFO");
                goto nolfb;
        }

        available=fix.smem_len;
        fprintf(stderr,"%dk video mem\r\n", available/1024);
        lfb=(unsigned char*)mmap(0, available, PROT_WRITE|PROT_READ, MAP_SHARED, fb, 0);
        if (!lfb)
        {
                perror("mmap");
                goto nolfb;
        }

        setmode(720, 576, 8);
        fprintf(stderr, "Framebuffer initialized. 720x576x8\n");
        
	for (i=0; i<256; i++)
        {
                d=i;
                d=(d-128)*(gamma+64)/(128+64)+128;
                d+=brightness-128; // brightness correction
                if (d<0)
                        d=0;
                if (d>255)
                        d=255;
                ramp[i]=d;
		// alpha correction
                if ( (255-(i*alpha/255)-96) >=0) 
                    rampalpha[i]=255-(i*alpha/255)-96;
                else rampalpha[i]=0;
        }

        return 1;
nolfb:
        lfb = 0;
        fprintf(stderr,"Framebuffer not available!\r\n");
        return -1;
}

int 
fb_set_cmap(unsigned char *buf, unsigned int start, unsigned int numcolors)
{
   struct fb_cmap cmap;
   
   int i;
   int d;
   
   fprintf(stderr, "Set CMAP: start, numcolors: %d, %d\n", (int) start, (int) numcolors);
   if ( (start>255) || (numcolors>256) ) 
   {
   	fprintf(stderr,"Start or numcolors was wrong.\n");
   	return 0;
   }
   d=0;
   for (i=0; i<(numcolors)*4*sizeof(CARD16) ; i+=8) 
   {
   	alpha[d] = rampalpha[buf[i+0]]<<8;//rampalpha[buf[i+0]]<<8;
   	red[d]   = ramp[buf[i+2]]<<8;
   	green[d] = ramp[buf[i+4]]<<8;
   	blue[d]  = ramp[buf[i+6]]<<8;
#if DEBUG
	fprintf(stderr, "Alpha %04X Red %04X, Green %04X, Blue %04X d=%d\n", alpha[d], red[d], green[d], blue[d], d); 
#endif
	d+=1;
   }
   cmap.start=start;
   cmap.len=numcolors;
   cmap.red=red;
   cmap.green=green;
   cmap.blue=blue;
   cmap.transp=alpha;
   
   if (ioctl(fb, FBIOPUTCMAP, &cmap)<0) 
   {
   	perror("FBIOPUTCMAP");
   }
   return 1;
}

void 
handle_remote_control(int sock)
{
    __u16 rccode;
    rfbKeyEventMsg msg;
    
    msg.type = rfbKeyEvent;
    
	if ( (read(rcfd, (char*) &rccode, 2) == 2)){
	   		fprintf(stderr, "RC: %04X\n", rccode);
	   		if ( (rccode&0xFF) != 0xFF ) {
	   			msg.key = rccode&0xFF;
	   			if ( (rccode&0x8000) == 0x8000)
	   				msg.down = 1;
	   			else 
	   				msg.down = 0;
	   			msg.pad = 0;
	   			tcp_putdata(sock, (char*) &msg, sz_rfbKeyEventMsg);
	   		}
	}
}

int
read_from_rfb_server(int sock, char *out, unsigned int n)
{
    int numbytes=0;
    int totalbytes=0;
 
        
    while (totalbytes<n)	
    {
	while ((numbytes = tcp_getdata(sock, out+totalbytes, n-totalbytes))==0)
	{
		if (OSDreceiver_active==0) return 0;
		handle_remote_control(sock);
	}
	totalbytes+=numbytes;
#if DEBUG
	fprintf(stderr,"Got %d bytes. Wanted %d bytes.\n",totalbytes, n);
        if (n<12) {
	   for (numbytes=0; numbytes<n; numbytes++)
	       fprintf(stderr,"%02X ",out[numbytes]);
	   fprintf(stderr,"\n");
        }    	
#endif
     }
	return 1;
}

int 
write_data_to_screen(unsigned int x, unsigned int y, unsigned int w, unsigned int h, char *buf)
{
   int i, j;
     
   fprintf(stderr, "write_data_to_screen: x,y,w,h: %d, %d, %d, %d\n", x, y, w, h);
   
   for (j=0; j<h; j++) 
      for (i=0; i<w; i++)
        lfb[x+i+((y+j)*720)] = buf[i+(j*w)] ;

   return 1;
}

int
handle_raw_encoded_message(rfbFramebufferUpdateRectHeader rectheader)
{
   unsigned long size;
   char *buf;
   size = (8/8 * (rectheader.r.w)) * (rectheader.r.h);  // 8/8 is bpp/8 (bytes per pixel)
   buf = malloc(size);
   if (!read_from_rfb_server(rfbsock, buf, size)) return 0;

   write_data_to_screen(
	 rectheader.r.x, 
	 rectheader.r.y, 
	 rectheader.r.w, 
	 rectheader.r.h, 
	 buf
	 );

   free (buf);
   return 1;
}

// OSD receiver thread
void 
*OSDreceiver(void *threadid)
{
   int size;
   char *buf;
   rfbServerToClientMsg msg;
   rfbFramebufferUpdateRectHeader rectheader;
   	
	OSDreceiver_active = 1;
	
	while (OSDreceiver_active) 
	{
	   if (!read_from_rfb_server(rfbsock, (char*)&msg, 1)) continue ;
	   switch (msg.type)
   	   {
   	      int i;
   	      case rfbFramebufferUpdate:
	   	      	 read_from_rfb_server(rfbsock, ((char*)&msg.fu)+1, sz_rfbFramebufferUpdateMsg-1);
	 		 msg.fu.nRects = Swap16IfLE(msg.fu.nRects);
	 		 for (i=0;i< msg.fu.nRects;i++)
	 		 {
	    		   read_from_rfb_server(rfbsock, (char*)&rectheader, sz_rfbFramebufferUpdateRectHeader);
	    		   rectheader.r.x = Swap16IfLE(rectheader.r.x);
	    		   rectheader.r.y = Swap16IfLE(rectheader.r.y);
	    		   rectheader.r.w = Swap16IfLE(rectheader.r.w);
	    		   rectheader.r.h = Swap16IfLE(rectheader.r.h);
	    		   rectheader.encoding = Swap32IfLE(rectheader.encoding);
	    		   switch (rectheader.encoding)
	    		   {
	       		     case rfbEncodingRaw:
		  	     	handle_raw_encoded_message(rectheader);		  
		  		break;
	       		     case rfbEncodingCopyRect:
		                fprintf(stderr, "Encoding CopyRect. Unimplemented.\n"); 
		   		break;
	       		     case rfbEncodingRRE:
		  	        fprintf(stderr, "Encoding RRE. Unimplemented.\n");
		  		break;
	       		     case rfbEncodingCoRRE:
		  	        fprintf(stderr, "Encoding CoRRE. Unimplemented.\n");
		  		break;
		             case rfbEncodingHextile:
				fprintf(stderr, "Encoding Hextile. Unimplemented.\n");
		  		break;
	       		     case rfbEncodingTight:
		  		fprintf(stderr, "Encoding Tight. Unimplemented.\n");
		  		break;
	       		     case rfbEncodingZlib:
		  		fprintf(stderr, "Encoding Zlib. Unimplemented.\n");
		  		break;
	       		     case rfbEncodingRichCursor:
		  		fprintf(stderr, "Encoding RichCursor. Unimplemented.\n");;
		  		break;
	       		     case rfbEncodingLastRect:
		  		printf("LAST\n");
		  		break;
	   	             default:
		  		printf("Unknown encoding\n");
		  		return 0;
		  		break;
	    		   }
	    		   
		 	}
			break;

      	      case rfbSetColourMapEntries:
	 	   	 fprintf(stderr, "SetColourMapEntries\n");
	 		 read_from_rfb_server(rfbsock, ((char*)&msg.scme)+1, sz_rfbSetColourMapEntriesMsg-1);
	 		 size = Swap16IfLE(msg.scme.nColours);
	 		 fprintf(stderr,"nColours: %d\n", size);
	 		 buf = malloc(sizeof(CARD16) * size * 4); // A, R, G, B 16-bit each 
	 		 read_from_rfb_server(rfbsock, buf, size * 4 * sizeof(CARD16));
	 		 fb_set_cmap(buf, Swap16IfLE(msg.scme.firstColour), size);
			 free(buf);
	 		 break;
      	      case rfbBell:
	 	 	 fprintf(stderr, "Bell message. Unimplemented.\n");
	 		 break;
      	      case rfbServerCutText:
	 		 fprintf(stderr, "ServerCutText. Unimplemented.\n");
	 		 read_from_rfb_server(rfbsock, ((char*)&msg.sct)+1, sz_rfbServerCutTextMsg-1);
	 		 size = Swap32IfLE(msg.sct.length);
	 		 buf = malloc(sizeof(char) * size);
	 		 read_from_rfb_server(rfbsock, buf, size);
		         buf[size]=0;
	 		 printf("%s\n", buf);
	 		 free(buf);
	 		 break;
      	      default:
	 		 printf("Unknown server message. Type: %i\n", msg.type);
	 		 return 0;
	 		 break;
   	   }
        }
           
	fprintf(stderr,"OSDreceiver thread terminated (id:%x).\r\n", (int) threadid);
	fflush(stderr);
	pthread_exit(NULL);	
}

void 
RCInit(void)
{
	rcfd=open("/dev/rawir2", O_RDONLY|O_NONBLOCK);
        if (rcfd<0)
        {
                perror("/dev/rawir2");
        }
}

void
StartOSDreceiver(void)
{
   
   int rc;
   
   fb_init();
   
   RCInit();
      
   fprintf(stderr,"Creating OSDreceiver thread (id: 0x02).\n");
   rc = pthread_create(&OSDreceiver_thread, NULL, OSDreceiver, (void *) 2);
   if (rc){
      fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
      exit(-1);
   }
}

void
StopOSDreceiver(void)
{
   int rc, status;
   
   OSDreceiver_active = 0;
   rc = pthread_join(OSDreceiver_thread, (void *)&status);
   if (rc)
   {
      printf("ERROR; return code from pthread_join() is %d\n", rc);
      exit(-1);
   }
   fprintf(stderr,"OSDreceiver thread joined.\r\n" );
   close(rcfd);
}

//http://www.llnl.gov/computing/tutorials/pthreads/
//http://lists.trolltech.com/qt-interest/2004-08/thread00706-0.html
//http://www.scit.wlv.ac.uk/cgi-bin/mansec?2+sigwait
